@@ -17,3 +17,6 @@ from . import gen_filename_fxns
 from . import gen_notebook_fxns
 from . import non_std_cfgs
 from . import eta
+from . import expl_vis
+from . import measure_utils
+from . import jupyter_colab_compat
\ No newline at end of file
