@@ -19,8 +19,12 @@ import skimage
 import skimage.measure as measure
 import skimage.io as skio
 
+from matplotlib.patches import Ellipse
+from matplotlib.lines import Line2D
+
 from .. import czd_utils
 from .. import pointgen
+from .. import measure_utils
 
 __all__ = ['mask_size_at_pt',
            'get_central_mask',
@@ -30,7 +34,8 @@ __all__ = ['mask_size_at_pt',
            'parse_properties',
            'auto_inc_contrast',
            'random_subimg_sample',
-           'MosImg']
+           'MosImg',
+           'IterableMosImg']
 
 # Various small functions that simplify larger functions below
 
@@ -65,7 +70,7 @@ def mask_size_at_pt(input_mask, coords):
 
 
 # function for retrieving mask at center of image, ~nearest to center of image
-def get_central_mask(results):
+def get_central_mask(results, verbose=True):
     """Return the largest central mask region array, if any,
         from threshholding or Detectron2 prediction results.
 
@@ -73,6 +78,10 @@ def get_central_mask(results):
     ----------
     results : Detectron2 Prediction or list[array]
         Detectron2 predictions or list of mask arrays for central mask.
+    verbose : bool, optional
+        Determines whether function prints error messages (e.g., "NO CENTRAL
+        GRAIN FOUND"). The default is True.
+        
 
     Returns
     -------
@@ -97,7 +106,8 @@ def get_central_mask(results):
 
     #does not search for masks if no zircons segmented from image
     if masks == []:
-        print('NO ZIRCON MASKS FOUND')
+        if verbose:
+            print('NO GRAIN MASKS FOUND')
         return mask_found_bool, []
 
     #gets central points for masks/image, number of masks created
@@ -121,18 +131,18 @@ def get_central_mask(results):
         mask_found_bool = True
 
     #extends search if mask not found at exact center
-    while mask_found_bool is False and pts.in_bounds:
-        pts.next_pt()
-        curr_x, curr_y = pts.curr_pts
-        for i in range(0, num_masks):
-            if masks[curr_y, curr_x, i]:
-                central_mask_indices.append(i)
-        if len(central_mask_indices) > 0:
-            mask_found_bool = True
-            break
+    for each_x, each_y in pts:
+        if mask_found_bool is False:
+            for i in range(0, num_masks):
+                if masks[each_y, each_x, i]:
+                    central_mask_indices.append(i)
+            if len(central_mask_indices) > 0:
+                mask_found_bool = True
+                break
 
     if not mask_found_bool:
-        print('ERROR: NO CENTRAL ZIRCON MASK FOUND')
+        if verbose:
+            print('ERROR: NO CENTRAL GRAIN MASK FOUND')
         return(mask_found_bool, [])
 
     #if only one mask at center, return it
@@ -178,7 +188,7 @@ def get_main_region_props(input_central_mask):
 
 def save_show_results_img(original_image, analys_name, display_bool=False,
                           save_dir='', tag_bool=False, input_central_mask=None,
-                          main_region=None, scale_factor=None):
+                          main_region=None, scale_factor=None, **kwargs):
     """Save and/or display a (optionally) scaled figure showing a shot-image,
        w/ or w/o mask and axes overlaid.
 
@@ -208,6 +218,13 @@ def save_show_results_img(original_image, analys_name, display_bool=False,
     scale_factor : float, optional
         Scale factor (microns/pixel) for original image and mask.
         The default is None.
+    **kwargs :
+        Args for plotting - fig_dpi = int; will set plot dpi to input integer.
+                            show_ellipse = bool; will plot ellipse corresponding
+                                           to maj, min axes if True.
+                            show_legend = bool; will plot a legend on plot if
+                                          True.
+                            
 
     Returns
     -------
@@ -243,10 +260,14 @@ def save_show_results_img(original_image, analys_name, display_bool=False,
     if not isinstance(input_central_mask, type(None)):
         if not isinstance(main_region, type(None)):
             overlay_bool = True
+    #check kwargs
+    plot_ellipse = False
+    if 'show_ellipse' in kwargs:
+        plot_ellipse = True
     if overlay_bool:
 
         ax.imshow(input_central_mask, alpha=0.4)
-
+        
         #plots measured axes atop image
         #for props in regions:
         y0, x0 = main_region.centroid
@@ -255,10 +276,37 @@ def save_show_results_img(original_image, analys_name, display_bool=False,
         y1 = y0 - math.sin(orientation) * 0.5 * main_region.minor_axis_length
         x2 = x0 - math.sin(orientation) * 0.5 * main_region.major_axis_length
         y2 = y0 - math.cos(orientation) * 0.5 * main_region.major_axis_length
+        if 'moment' in str(main_region.best_ax_from):
+            ax.plot((x0, x1), (y0, y1), '-r', linewidth=1.5,
+                    label = 'Major, minor axes')
+            ax.plot((x0, x2), (y0, y2), '-r', linewidth=1.5)
+        else:
+            ax.plot((x0, x1), (y0, y1), '--r', linewidth=1.5,
+                    label = 'Major, minor axes', alpha =0.5,)
+            ax.plot((x0, x2), (y0, y2), '--r', linewidth=1.5, alpha =0.5)
+        ax.plot(x0, y0, '.g', markersize=10, label = 'Centroid')
+        
 
-        ax.plot((x0, x1), (y0, y1), '-r', linewidth=1.5)
-        ax.plot((x0, x2), (y0, y2), '-r', linewidth=1.5)
-        ax.plot(x0, y0, '.g', markersize=10)
+
+        #plot minimum bounding rectangle atop image (source of Feret diameter)
+        if plot_ellipse is True:
+            el_rot_deg = math.degrees(orientation) * -1.0
+            el = Ellipse((x0, y0), main_region.minor_axis_length,
+                          main_region.major_axis_length, angle=el_rot_deg,
+                          fill = False, linestyle='--', color='red', 
+                          alpha=0.5, linewidth=1.5)
+            ax.add_artist(el)
+
+
+        #plot minimum bounding rectangle atop image (source of Feret diameter)
+        if 'rect' in str(main_region.best_ax_from):
+            ax.plot(*main_region.rect_points, color = 'b',
+                    linestyle = '-', alpha =0.7, linewidth=1.0,
+                    label = 'Minimum area rectangle')
+        else:
+            ax.plot(*main_region.rect_points, color = 'b',
+                    linestyle = '--', alpha =0.35, linewidth=1.0,
+                    label = 'Minimum area rectangle')
 
     #mark 'failed' shots
     else:
@@ -269,20 +317,33 @@ def save_show_results_img(original_image, analys_name, display_bool=False,
     ax.set_yticks(y_tick_locs)
     ax.set_yticklabels([str(label) for label in y_tick_labels])
 
+    #additional kwarg plotting options
+    if 'show_legend' in kwargs:
+        if kwargs['show_legend'] is True:
+            handles, labels = ax.get_legend_handles_labels()
+            if plot_ellipse is True:
+                handles, labels = ax.get_legend_handles_labels()
+                handles.append(Line2D([0], [0], linestyle='--', color='red',
+                                      alpha=0.5, linewidth=1.5))
+                labels.append("Ellipse with same $2_{nd}$ order\nmoments as grain mask")
+            ax.legend(handles=handles, labels = labels)
+    if 'fig_dpi' in kwargs:
+        fig.set_dpi(int(kwargs['fig_dpi']))
+
     if save_dir:
         img_save_filename = os.path.join(save_dir, adj_analys_name + '.png')
         if tag_bool:
             img_save_filename = os.path.join(save_dir,
                                              adj_analys_name + '_tagged.png')
 
-        plt.savefig(img_save_filename)
+        fig.savefig(img_save_filename)
 
     #plt.clf()
 
     if display_bool:
         plt.show()
 
-    plt.close('all')
+    plt.close(fig)
 
 
 #measures zircon mask (in pixels) using skimage, creates an image by \
@@ -290,7 +351,7 @@ def save_show_results_img(original_image, analys_name, display_bool=False,
 # (if display_bool = True) and/or saves (if save_dir != '') this image.
 def overlay_mask_and_get_props(input_central_mask, original_image, analys_name,
                                display_bool=False, save_dir='', tag_bool=False,
-                               scale_factor=None):
+                               scale_factor=None, **kwargs):
     """Measures zircon mask (in pixels) using skimage, creates an image by
        overlaying the mask atop the original subimage, and optionally displays
        (if display_bool = True) and/or saves (if save_dir != '') this image.
@@ -316,6 +377,15 @@ def overlay_mask_and_get_props(input_central_mask, original_image, analys_name,
     scale_factor : float, optional
         Scale factor (microns/pixel) for original image and mask.
         The default is None.
+    **kwargs :
+        Plotting-related kwargs, passed in full to czd_utils.save_show_results_img.
+            fig_dpi = int; will set plot dpi to input integer.
+            show_ellipse = bool; will plot ellipse corresponding
+                           to maj, min axes if True.
+            show_box = bool; will plot the minimum area rect.
+                       if True.
+            show_legend = bool; will plot a legend on plot if
+                          True.
 
     Returns
     -------
@@ -328,12 +398,12 @@ def overlay_mask_and_get_props(input_central_mask, original_image, analys_name,
     input_central_mask = czd_utils.mask_to_3D_arr_size(input_central_mask,
                                                        original_image)
     #get main region properties
-    main_region = get_main_region_props(input_central_mask)
+    main_region = measure_utils.box_main_region_props(input_central_mask)
 
     if display_bool or save_dir:
         save_show_results_img(original_image, analys_name, display_bool,
                               save_dir, tag_bool, input_central_mask,
-                              main_region, scale_factor)
+                              main_region, scale_factor, **kwargs)
 
     return main_region
 
@@ -366,7 +436,7 @@ def parse_properties(props, img_scale_factor, analys_name, verbose = False,
 
     """
     #initialize empty (null values) props list
-    props_list = [analys_name, 0, 0, 0, 0, 0, 0, 0, 0, img_scale_factor]
+    props_list = [analys_name, *czd_utils.get_save_fields(get_nulls=True), img_scale_factor]
     #if no region found, skip and return null properties
     if props == []:
         if verbose:
@@ -383,16 +453,23 @@ def parse_properties(props, img_scale_factor, analys_name, verbose = False,
     major_leng = props.major_axis_length * img_scale_factor
     minor_leng = props.minor_axis_length * img_scale_factor
     roundness = 4 * math.pi * props.area / props.perimeter**2
+    Feret_diam = props.rect_major_axis_length * img_scale_factor
+    orth_feret_diam = props.rect_minor_axis_length * img_scale_factor
+    best_major_ax = props.best_major_ax_length * img_scale_factor
+    best_minor_ax = props.best_minor_ax_length * img_scale_factor
+    best_ax_from = props.best_ax_from
     scale_factor = img_scale_factor
 
     props_list = [analys_name, area, convex_area, eccent, eq_diam, perim,
-                  major_leng, minor_leng, roundness, scale_factor]
+                  major_leng, minor_leng, roundness, Feret_diam,
+                  orth_feret_diam, best_major_ax, best_minor_ax,
+                  best_ax_from, scale_factor]
     for addit_prop in addit_props:
         props_list.append(addit_prop)
 
     if verbose:
-        print('Major axis length =', round(major_leng, 1), 'µm,',
-              'Minor axis length =', round(minor_leng, 1), 'µm')
+        print('Major axis length =', round(best_major_ax, 1), 'µm,',
+              'Minor axis length =', round(best_minor_ax, 1), 'µm')
 
     return props_list
 
@@ -421,13 +498,13 @@ def auto_inc_contrast(input_image, low_cont_threshhold =0.10):
 
         ## code for enhancing contrast by channel: \
         ## avoids warning but produces grainy images
-        #red = skimage.exposure.equalize_hist(input_image[:, :, 0])
-        #green = skimage.exposure.equalize_hist(input_image[:, :, 1])
-        #blue = skimage.exposure.equalize_hist(input_image[:, :, 2])
-        #return skimage.img_as_ubyte(np.stack([red, green, blue], axis=2))
+        red = skimage.exposure.equalize_hist(input_image[:, :, 0])
+        green = skimage.exposure.equalize_hist(input_image[:, :, 1])
+        blue = skimage.exposure.equalize_hist(input_image[:, :, 2])
+        return skimage.img_as_ubyte(np.stack([red, green, blue], axis=2))
 
         #code for enhancing contrast without splitting channels: results are less grainy
-        return skimage.img_as_ubyte(skimage.exposure.equalize_hist(input_image))
+        #return skimage.img_as_ubyte(skimage.exposure.equalize_hist(input_image))
     else:
         return input_image
 
@@ -543,9 +620,9 @@ class MosImg:
             #calculates a new scale factor (microns/pixel) from .Align file
             self.scale_factor = czd_utils.calc_scale_factor(self.align_data[2:],
                                                             self.mos_dims[1:])
-            #converts shifts to pixels from microns
-            self.pix_shift_list = [val/self.scale_factor
-                                   for val in x_y_shift_list]
+            #initialize shift list (in microns)
+            self.pix_shift_list = [val for val in x_y_shift_list]
+            
             #calculates origin based on new variables
             self.x_origin = float(self.align_data[0] - self.align_data[2]/2
                                   - self.pix_shift_list[0])
@@ -571,7 +648,7 @@ class MosImg:
         Parameters
         ----------
         new_sub_img_size : int
-            New size for subimages.
+            New size for subimages, in microns.
 
         Returns
         -------
@@ -598,9 +675,17 @@ class MosImg:
             A list of pixel coordinates [x pixel coord, y pixel coord].
 
         """
-
-        return [round(float((x_coord - self.x_origin) / self.scale_factor)),
-                round(float((y_coord - self.y_origin) / self.scale_factor))]
+        #unrotated mapping if no rotation in align file
+        if float(self.align_data[4]) == 0.0:
+            return [round(float((x_coord - self.x_origin) / self.scale_factor)),
+                    round(float((y_coord - self.y_origin) / self.scale_factor))]
+        else:
+            #back-rotate coordinates by rotation given in .Align file
+            x_coord, y_coord = czd_utils.rotate_pt((x_coord, y_coord),
+                                                   self.align_data[4],
+                                                   self.align_data[:2])
+            return [round(float((x_coord - self.x_origin) / self.scale_factor)),
+                    round(float((y_coord - self.y_origin) / self.scale_factor))]
 
     def set_subimg(self, x_coord, y_coord):
         """Set a clipped subimage centered on entered coordinates.
@@ -657,3 +742,164 @@ class MosImg:
                      self.x_y_0_offsets[0]:self.x_y_1_offsets[0],
                      :] = self.full_img[self.x_y_0[1]:self.x_y_1[1],
                                         self.x_y_0[0]:self.x_y_1[0], :]
+    def get_subimg(self, *coords_inpt):
+        """Get the current sub_img, after optionally adjusting the current
+        sub_img location to input coordinates if they are given.
+
+        Parameters
+        ----------
+        coords_inpt : 2 ints and/or floats, optional
+            If given, they will be treated as x_coord, y_coord and used to set
+            a new sub_img location. Otherwise, location will be kept as-is.
+
+        Returns
+        -------
+        np.ndarray
+            Array representing the current subimage (i.e., self.sub_img).
+
+        """
+        if coords_inpt:
+            self.set_subimg(*coords_inpt)
+        return self.sub_img
+    
+    def get_zoomed_subimg(self, zoom_fact=1.1):
+        """Get a version of the current subimage that is zoomed in or out by
+        the given zoom_fact (i.e., decreasing or increasing spatial extent on
+        the mosaic image, keeping the same center). Reset instance sub_img_size
+        afterwards.
+
+        Parameters
+        ----------
+        zoom_fact : int or float, optional
+            Factor to multiply the current sub_img extent by. The default is
+            1.1.
+
+        Returns
+        -------
+        out_img: np.ndarray
+            Array representing the current subimage (i.e., self.sub_img), with
+            extent adjusted by zoom_fact.
+
+        """
+        self.set_sub_img_size(self.sub_img_size*zoom_fact)
+        out_img = self.get_subimg()
+        self.set_sub_img_size(self.sub_img_size_input)
+        return out_img
+
+
+class IterableMosImg:
+    """Creates and links a MosImage instance to a scan dict loaded from a
+    corresponding dict loaded from a .scancsv, with scan coordinates. Iterable
+    for easy use with parallel processing.
+
+    Parameters
+    ----------
+    mosaic_image_path : str
+        Full file path to a .bmp mosaic image file.
+    scan_dict : dict{Any: [float|int, float|int]}
+        A dictionary with scan names: scan coordinates, as loaded from a
+        .scancsv file using czd_utils.scancsv_to_dict().
+    try_bools : list[bool], optional
+        A list of bools corresponding to alternate methods (scale jittering,
+        contrast enhancement, and/or Otsu thresholding) to iteratively try on
+        a scan image if inital central grain segmentation is unsuccesful.
+        Format: [Try_zoomed_out_subimage, Try_zoomed_in_subimage,
+                 Try_contrast_enhanced_subimage,
+                 Try_Otsu_thresholding].
+        The default is [False, False, False, False].
+    align_file_path : str, optional
+        Path to a .Align file corresponding to the mosaic image.
+        If not entered, the image scale will be in pixels.
+        The default is ''.
+    sub_img_size : int, optional
+        Size of subimages extracted from the mosaic in pixels or microns.
+        The default is 500.
+    x_y_shift_list : list[int], optional
+        A list [x shift, y shift] used to correct misalignment between
+        .scancsv shot coordinates and a mosaic image. The default is [0, 0].
+
+    Returns
+    -------
+    None.
+
+    """
+    
+    def __init__(self, mosaic_image_path, scan_dict,
+                 try_bools=[False, False, False, False],
+                 align_file_path = '',
+                 sub_img_size = 500, x_y_shift_list = [0, 0]):
+
+        #create the child mosaic image instance
+        self._mos_img = MosImg(mosaic_image_path, 
+                               align_file_path=align_file_path,
+                               sub_img_size=sub_img_size,
+                               x_y_shift_list=x_y_shift_list)
+        self._scan_dict = scan_dict
+        self.try_bools = try_bools
+
+    def get_aug_subimgs(self, *inpt_pt_coords):
+        """Get a list of subimgs (as np.ndarrays) from the child MosImg. These
+        will depend on the first three try bools.
+
+        Parameters
+        ----------
+        *inpt_pt_coords : float|int, float|int, optional
+            Coordinates to set subimage location. If not included, subimages
+            will be drawn from the child MosImg current subimg location.
+
+        Returns
+        -------
+        out_imgs_lst : list[np.ndarray]
+            Images for segmentation. These depend on the first two booleans in
+            self.try_bools. The number of returned subimages in the list will
+            thus range from 1 to 3. The first subimage will always be the
+            unaltered (i.e., 'original') subimage clipped from the child
+            MosImg instance, at the current target location.
+            Format:
+                [unaltered subimage at target location,
+                 if try_bools[0]: subimage with extent increased w/ fact. of 1.1,
+                 if try_bools[0]: subimage with extent decreased w/ fact. of 0.9,]
+
+        """
+        if inpt_pt_coords:
+            self._mos_img.set_subimg(*inpt_pt_coords)
+        out_imgs_lst = [self._mos_img.get_subimg()]
+        #if first try bool is true, we need a 'zoomed-out' subimage (extent*1.1)
+        if self.try_bools[0]:
+            out_imgs_lst.append(self._mos_img.get_zoomed_subimg(1.1))
+        #if second is True, we need a 'zoomed-in' subimage (extent*0.9)
+        if self.try_bools[1]:
+            out_imgs_lst.append(self._mos_img.get_zoomed_subimg(0.9))
+        return out_imgs_lst
+
+    def __iter__(self):
+        """Iterate through the scan dict, yielding all scan-dependent variables
+        needed for segmentation and central grain dimensional analysis of the
+        target grain.
+
+        Yields
+        ------
+        idx_count : int
+            Current index-location within the scan dict.
+        str
+            Current scan name (from .scancsv file, by way of the scan dict).
+        out_imgs : list[np.ndarray]
+            List of images (as np.ndarrays).These depend on the first two booleans in
+            self.try_bools. The number of returned subimages in the list will
+            thus range from 1 to 3. The first subimage will always be the
+            unaltered (i.e., 'original') subimage clipped from the child
+            MosImg instance, at the current target location.
+            Format:
+                [unaltered subimage at target location,
+                 if try_bools[0]: subimage with extent increased w/ fact. of 1.1,
+                 if try_bools[0]: subimage with extent decreased w/ fact. of 0.9,]
+        out_scale_factor : float
+            Scale in um/pixel for spot. From loaded child MosImg instance.
+
+        """
+        idx_count = 0
+        for eachscan, eachscan_coords in self._scan_dict.items():
+            out_imgs = self.get_aug_subimgs(*eachscan_coords)
+            out_scale_factor=self._mos_img.scale_factor
+            yield (idx_count,str(eachscan), out_imgs, out_scale_factor)
+            idx_count +=1
